/* C-SLang 2.0. samplesvirm.c
   This file umplements the self-test engine:
   Based on the table of tests, it configures the
   virtual machine (SVIRM) as requested by the test
   and runs the test's script both compiled-in and exported
   to bytecode and also in a run and in a single-step execution.
   
   (C) 2003 MacroExpressions.com 

   Version control:
	$Date: 2004/08/07 $
	$Change: 70 $
	$Revision: #6 $
*/

#include <stdio.h> /*for printf */
#include <string.h>
#include <limits.h>
#include "cslpubl.h"
#include "sample.h"

/* Null-terminated Table Of Tests */
const sample_t *samples[] = {
    &arithm_svirmdata,
    &Charcount_svirmdata,
    &SAE_example_svirmdata,
    NULL
};


/* This should be enough for all
 */
static tempvar_t temp[256]; 
static inpvar_t inp[256];
static outpvar_t outp[256];
static CtlStack_t cstack[15];

/* This is an array to hold C-SLang bytecode in exported format */
static unsigned char myBytecode[2004];

/* sample breakpoints */
static CSLbp_t mybp[] = {
    {NULL, 1, 6},
};

/* This is bytecode emit (export) function. In this case, it simply fills the
 * char array
 */
static int mycount;
static void mydump(int c, void *arg)
{
	(void)arg; /*OK unused*/
    myBytecode[mycount++] = (unsigned char)c;
}

static void log_result(FILE *log, const sample_t *test)
{
    int i;
    if(test->num_temps) {
        fprintf(log, "       Temporaries: ");
        for(i=0; i<test->num_temps; i++) {
            fprintf(log, " %04X", temp[i]);
        }
        fprintf(log, "\n");
    }
    if(test->num_outps) {
        fprintf(log, "       Outputs: ");
        for(i=0; i<test->num_outps; i++) {
            fprintf(log, " %02X", outp[i]);
        }
        fprintf(log, "\n");
    }
    fprintf(log, "\n");
}

static void log_internals(FILE *log, CSLangRun_t *run)
{
    fprintf(log, "       function %d, instruction %d:\n", run->func_id, run->num_instr);
    fprintf(log, "       A=%08X X=%08X Call depth = %d\n",
        run->Accumulator, run->IndexRegister, run->CallDepth);
}

static void inittest0(FILE *log, SVIRM_t *svirm, const sample_t *test)
{
    int i;

    /*Fill the arrays*/
    memset(temp, -1, sizeof(temp));
    memset(outp, -1, sizeof(outp));
    memset(inp, -1, sizeof(inp));

    /* Fill the svirm */
    svirm->infunc = test->infunc;
    svirm->outfunc = test->outfunc;
    svirm->inp.buf = inp;
    svirm->inp.size = test->inputs.size;
    svirm->outp.buf = outp;
    svirm->outp.size = test->num_outps;
    svirm->tempbuf.buf = temp;
    svirm->tempbuf.size = test->num_temps;

    /* Copy inputs */
    memcpy(inp, test->inputs.buf, test->inputs.size*sizeof(*inp));

    /* Output the inputs supplied */
    fprintf(log, "  Inputs:");
    for(i=0; i<svirm->inp.size; i++) {
        fprintf(log, " %02X", svirm->inp.buf[i]);
    }
    fprintf(log, "\n");
}

static void inittest1(FILE *log, CSLangRun_t *run, int ret)
{
    /* Initialize the run */
    if(ret == 0) {
        fprintf(log, "  Initialized script\n");
        CSLangSetControlStack(run, cstack, nelem(cstack));
        SCLangSetBreakpoints(run, mybp);
    }
    else {
        fprintf(log, "  initialization error %d\n", ret);
    }

}

static int inittest_com(FILE *log, SVIRM_t *svirm, CSLangRun_t *run,
                        const sample_t *test, CSLscript_t script)
    
{
    int ret;
    inittest0(log, svirm, test);
    ret = CSLangInitCom(run, script, svirm);
    inittest1(log, run, ret);
    return ret;
}

static int inittest_exp(FILE *log, SVIRM_t *svirm, CSLangRun_t *run,
                        const sample_t *test, const unsigned char *script)
    
{
    int ret;
    inittest0(log, svirm, test);
    ret = CSLangInitExp(run, script, svirm);
    inittest1(log, run, ret);
    return ret;
}



static int exectest(FILE *log, CSLangRun_t *run, const sample_t *test)
{
    int ret, step_count;

    step_count = 0;
    do {
        ret = CSLangExec(run ); /*call the first registered function in script*/
        if(ret < 0) {
            fprintf(log, "Error %d at ", ret);
            log_internals(log, run);
            break;
        }
        else if(ret == CSL_BREAKPOINT) {
            fprintf(log, "         Hit breakpoint at ");
            log_internals(log, run);
            log_result(log, test);
        }
        else if(ret == CSL_STEP_COMPLETED) {
            step_count++;
        }
    } while(ret!=CSL_DONE);
    
    /* Output results */
    fprintf(log, "     Completed with code %d\n", ret);
    if(run->single_step) {
        fprintf(log, "      Stopped after %d instructions\n", step_count);
    }
    log_result(log, test);
    if(ret == CSL_DONE) {
        ret = 0;
    }
    return ret;
}

static int runtest(FILE *log, const sample_t *test)
{
    SVIRM_t svirm;
    CSLangRun_t run;
    int ret, err;

    /* Announce */
    fprintf(log, "\n\nRunning \"%s\":\n", test->name);

    fprintf(log, " 1. Compiled-in all at once:\n");
    ret = err = inittest_com(log, &svirm, &run, test, *test->script);
    if(err == 0) {
        err = exectest(log, &run, test);
    }
    if(!ret && err) ret = err;
    
    fprintf(log, " 2. Compiled-in single-step:\n");
    err = inittest_com(log, &svirm, &run, test, *test->script);
    if(err == 0) {
        SCLangSetSingleStep(&run, 1);
        err = exectest(log, &run, test);
    }
    if(!ret && err) ret = err;

    fprintf(log, " 3. Exporting to bytecode\n");
    mycount = 0;
    err = CSLangBytecodeExport(*test->script, mydump, NULL);
    if(err == 0) {
        if(((myBytecode[0]<<CHAR_BIT)|myBytecode[1]) == mycount) {
            fprintf(log, "   Bytecode length = %d\n", mycount);
        }
        else {
            fprintf(log, "   Error: length mismatch\n");
            err = -1000;
        }
    }
    if(err) {
        return err;
    }

    fprintf(log, " 4. Bytecode all at once:\n");
    err = inittest_exp(log, &svirm, &run, test, myBytecode);
    if(err == 0) {
        SCLangSetSingleStep(&run, 0);
        err = exectest(log, &run, test);
    }
    if(!ret && err) ret = err;

    fprintf(log, " 5. Bytecode single-step:\n");
    err = inittest_exp(log, &svirm, &run, test, myBytecode);
    if(err == 0) {
        SCLangSetSingleStep(&run, 1);
        err = exectest(log, &run, test);
    }
    if(!ret && err) ret = err;

    return ret;
}

void test_all(FILE *log) {
    const sample_t **test;
    for(test=samples; *test != NULL; test++) {
        runtest(log, *test);
    }
}
